import numpy as np
import scipy.special as sc
from scipy.optimize import fsolve
import matplotlib.pyplot as plt


class BattjesGroenendijk:
    """Formulas of Battjes and Groenendijk (2000)

    The wave height distributions on shallow foreshores deviates from
    those in deep water due to the effects of the restricted
    depth-to-height ratio and of wave breaking. Battjes and Groenendijk
    (2000), therefore derived a generalised empirical parameterisations
    based on laboratory data to determine these effects. This allows for
    the computation of all statistical wave heights based on the spectral
    wave height, water depth and slope of the foreshore.

    Parameters
    ----------
    Hm0 : float
        the spectral wave height [m]
    h : float
        the water depth [m]
    slope_foreshore : float or tuple
        the slope of the foreshore [rad], or as tuple formatted as (V,H)

    Attributes
    ----------
    Hrms : float
        the root-mean-square wave height [m]
    Htr_tilde : float
        the transitional wave height, made dimensionless with Hrms [m]
    k1, k2 : floats
        shape parameters of the Composite Weibull Distribution (CWD) [-]
    """

    def __init__(self, Hm0, h, slope_foreshore):
        """ See help(BattjesGroenendijk) for more info """
        # check if slope foreshore is given as a tuple (V, H)
        if isinstance(slope_foreshore, tuple):
            # if so, compute the angle
            slope_foreshore = np.arctan(slope_foreshore[0]/slope_foreshore[1])

        Htr = (0.35 + 5.8*np.tan(slope_foreshore))*h
        self.Hrms = (0.6725 + 0.2025*(Hm0/h))*Hm0
        self.Htr_tilde = Htr/self.Hrms

        self.k1 = 2
        self.k2 = 3.6

    @staticmethod
    def gammainc_upper(a, x):
        """Upper incomplete gamma function

        Defined as

        .. math::
            \Gamma(a,x) = \int_x^\infty t^{a - 1}e^{-t} dt

        Implemented using scipy.special as gamma(a)*gammaincc(a,x)

        Parameters
        ----------
        a : float
            positive parameter
        x : float
            nonnegative argument

        Returns
        -------
        float
            value of the upper incomplete gamma function
        """
        if a >= 0:
            out = sc.gamma(a)*sc.gammaincc(a, x)
            return out
        else:
            return None

    @staticmethod
    def gammainc_lower(a, x):
        """Lower incomplete gamma function

        Defined as

        .. math::
            \Gamma(a,x) = \int_0^x t^{a - 1}e^{-t} dt

        Implemented using scipy.special as gamma(a)*gammainc(a,x)

        Parameters
        ----------
        a : float
            positive parameter
        x : float
            nonnegative argument

        Returns
        -------
            float
                value of the lower incomplete gamma function
        """
        if a >= 0:
            out = sc.gamma(a)*sc.gammainc(a, x)
            return out
        else:
            return None

    def _solver(self, x):
        """Computes H1~ and H2~ for a given Htr~

        Computes H1~ and H2~ by solving equation 7 from Battjes and
        Groenendijk (2000), and the continuity condition for the
        Composite Weibull Distribution. This continuity conditions is:
        .. math::
            \frac{H_tr}{H_1}^{k_1} = \frac{H_tr}{H_2}^{k_2}

        The values :math:`H_{tr}`, :math:`H_{1}` and :math:`H_{2}` are
        then made nondimensional by dividing them by :math:`H_{rms}`. A
        tilde is added to the names of the values to indicate that they
        are the nondimensional values.

        Parameters
        ----------
        x : list
            initial guess for the values of H1~ and H2~, [float, float]

        Returns
        -------
        list
            containing the continuity equation and equation 7 of
            Battjes and Groenendijk (2000) used to find the values of
            H1~ and H2~ for a given Htr~
        """
        a1 = 2/self.k1 + 1
        a2 = 2/self.k2 + 1

        eq_continuity = (self.Htr_tilde
                         * (self.Htr_tilde/x[0])**(-self.k1/self.k2) - x[1])

        eq7 = (np.sqrt(x[0]**2
                       * self.gammainc_lower(a1,(self.Htr_tilde/x[0])**self.k1)
                       + x[1]**2
                       * self.gammainc_upper(a2,(self.Htr_tilde/x[1])**self.k2))
               - 1)

        return [eq_continuity, eq7]

    def get_Hp(self, P):
        """Computes :math:`H_{P}`

        This method computes the wave height exceeded by P% of the
        waves. It uses scipy.optimize.fsolve to determine H1~ and H2~,
        which are then used to compute :math:`H_{P}`

        Parameters
        ----------
        P : float
            exceedance probability as a fractal [-]

        Returns
        -------
        H_P : float
            wave height exceeded by P\\% of the waves

        Raises
        ------
        ValueError
            If P is entered as a percentage, must be entered as P%/100
        """
        if P >= 1:
            raise ValueError('Do not enter P as a percentage, but P%/100')
        H1_H2 = fsolve(self._solver, [1, 1])

        H_CWD_1 = H1_H2[0]*(-np.log(P))**(1/self.k1)

        if H_CWD_1 <= self.Htr_tilde:
            return H_CWD_1*self.Hrms
        else:
            H_CWD_2 = H1_H2[1]*(-np.log(P))**(1/self.k2)
            return H_CWD_2*self.Hrms

    def get_Hn(self, N):
        """Computes :math:`H_{1/N}`

        This method computes the mean of the highest 1/N-part of the
        wave heights. It uses scipy.optimize.fsolve to determine H1~
        and H2~, which are then used to compute :math:`H_{1/N}`.

        Parameters
        ----------
        N : float
            highest 1/N-part of the wave heights [-]

        Returns
        -------
        H_1/N : float
            mean of the highest 1/N part of the waves

        Raises
        ------
        ValueError
            if N is smaller than 1
        """
        if N >= 1:
            H1_H2 = fsolve(self._solver, [1, 1])

            H1 = H1_H2[0]
            H2 = H1_H2[1]

            a1 = 1/self.k1 + 1
            a2 = 1/self.k2 + 1

            H_N = H1*(np.log(N))**(1/self.k1)

            if self.Htr_tilde >= H_N:
                H_1 = (N*H1*(self.gammainc_upper(a1, np.log(N))
                       - self.gammainc_upper(a1, (self.Htr_tilde/H1)**self.k1))
                       + N*H2
                       * self.gammainc_upper(a2, (self.Htr_tilde/H2)**self.k2))

                return H_1*self.Hrms
            else:
                H_2 = N*H2*self.gammainc_upper(a2, np.log(N))

                return H_2*self.Hrms
        else:
            msg = f'N = {N} has no meaning, please enter a value N >= 1'
            raise ValueError(msg)



from ..utils.exceptions import user_warning

def scour_protection(L, slope=None):
    """ Compute the length of the scour protection

    Compute the estimated length of the scour hole with Sumer and
    Fredsoe (2000). In the table the estimated length of the scour hole
    from the experimental study is presented.

    +-------------------------+------------------+
    | slope of the breakwater | estimated length |
    +=========================+==================+
    | Vertical breakwater     | 1.0(L/4)         |
    +-------------------------+------------------+
    | 1:1.2                   | 0.6(L/4)         |
    +-------------------------+------------------+
    | 1:1.75                  | 0.4(L/4)         |
    +-------------------------+------------------+

    Parameters
    ----------
    L : float
        wave length at the toe of the structure, computed with the mean
        wave period [m]
    slope : float, optional, default: None
        Slope of the armour layer (V, H), for example a slope of 3V:4H
        is defined as (3, 4). By default the length of
        the scour protection for a vertical breakwater is computed.

    Returns
    -------
    float
        the required length of the scour protection
    """
    # check if a slope is given
    if slope is None:
        # no slope, thus vertical breakwater
        W = L/4

    else:
        # slope is given
        slope_div = slope[0]/slope[1]

        # determine scalar for estimating the length
        scalar = 0.4 + 0.2 * ((slope_div - 1/1.75)/(1/1.2 - 1/1.75))

        # compute length of the scour hole
        W = scalar*(L/4)

        if slope_div >= 1/1.75 and slope_div <= 1/1.2:
            # in range of experiments
            pass
        else:
            # slope is out of range, so print a warning
            user_warning(
                (f'slope {slope[0]}:{round(slope[1],2)} is out of range in '
                  'scour_protection(), must be between 1:1.75 and 1:1.2'))

    return W



import context
import unittest
import warnings
import numpy as np

from breakwater.design import Configurations
from breakwater.conditions import LimitState
from breakwater.material import RockGrading, Xbloc


class TestCombinations(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # define the config file, with the input
        # of this input all possible combinations are generated
        cls.config = {
            'A': None,
            'B': np.linspace(1, 5, 5),
            'C': np.linspace(11, 15, 5),
            'D': 18,
            'E': np.linspace(21, 25, 5)}

        # define vkwargs to set varying and fixed parameters
        vkwargs = {
            'A': {'Constant': True},
            'B': {'Constant': False},
            'C': {'Constant': False},
            'D': {'Constant': True},
            'E': {'Constant': False}}

        # generate combinations
        cls.varying, cls.max_combinations = Configurations._get_combinations(
            vkwargs=vkwargs, config=cls.config)

    def test_max_combinations(self):
        # check maximum number of combinations
        self.assertEqual(self.max_combinations, 125)

    def test_check_combinations(self):
        # check if A and D are not in the varying dict
        self.assertNotIn('A', self.varying)
        self.assertNotIn('D', self.varying)

        # define index for B, C and E
        i_B, i_C, i_E = 0, 0, 0

        # check combinations
        for i in range(self.max_combinations):
            # get unique combination
            combination = Configurations._get_concept_set(
                configs=self.varying, id=i+1)

            # check combination
            self.assertEqual(combination['B'], self.config['B'][i_B])
            self.assertEqual(combination['C'], self.config['C'][i_C])
            self.assertEqual(combination['E'], self.config['E'][i_E])

            # increment E index every combination
            i_E += 1

            # increment C index every 5 combinations
            if (i+1) % 5 == 0:
                i_C += 1

                # reset E index
                i_E = 0

            # increment B index every 25 combinations
            if (i+1) % 25 == 0:
                i_B += 1

                # reset C and E index
                i_C, i_E = 0, 0


class TestConfigurations(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        ULS = LimitState(
            h=15, Nod=0.5, Sd=2, Ho=5, q=10, Hm0=3.22, T_m_min_1=8.66,
            label='ULS', Hmax=6)
        SLS = LimitState(
            h=10, Nod=0.5, Sd=2, Ho=4, q=20, Hm0=2.52, T_m_min_1=7.66,
            label='SLS')

        with warnings.catch_warnings(record=True) as w:
            ULS.transform_periods(0.5)
            SLS.transform_periods(0.5)
            ULS.check_deep_water()
            SLS.check_deep_water()

        xbloc = Xbloc()
        NEN = RockGrading()

        cls.configs = Configurations(
            structure=['RRM', 'CRM', 'RC', 'CC'], LimitState=ULS,
            slope=((1,6), (3,4), 6), slope_foreshore=(1,100), rho_w=1000,
            B=(3,8,5), N=3000, P=0.4, Grading=NEN, Dn50_core=(0.1, 0.4, 5),
            safety=1, slope_toe=(2,3), ArmourUnit=xbloc, rho_c=2400,
            rho_fill=1600, Pc=(0.6, 0.8, 5), Bm=(4,8,5),  hb=(1.5,5,5),
            mu=0.5, SF_sliding=1.2, SF_turning=1.2, beta=0,
            slope_foundation=(2,3), BermMaterial=xbloc)

    def test_values(self):
        # test parameters and values in the df
        # define list with columns to exclude as these are not parameters
        exclude = ['concept', 'id', 'type', 'warnings']

        # drop columns in exclude and define new df
        df = self.configs.df.drop(columns=exclude)

        # defined values
        defined = {
            'B': [3, 4.25, 5.5, 6.75, 8, np.nan],
            'B_toe': [None, np.nan],
            'Bm': [np.nan, 4, 5, 6, 7, 8],
            'Dn50_core': [0.1, 0.175, 0.25, 0.325, 0.4, np.nan],
            'Pc': [np.nan, 0.6, 0.65, 0.7, 0.75, 0.8],
            'hb': [np.nan, 1.5, 2.375, 3.25, 4.125, 5],
            'slope': [9.46, 15.82, 21.8, 27.32, 32.35, 36.87, np.nan],
            'slope_foundation': [np.nan, 33.69],
            'slope_toe': [33.69, np.nan]}

        # iterate over the columns of the df of the parameters
        for parameter in df.columns:
            # check if parameter is slope
            if 'slope' in parameter:
                # check if not all nan
                if not df[parameter].isnull().all():
                    # convert tuples to floats
                    df[parameter] = df[parameter].apply(
                        lambda x: convert(x))

            # get unique and correct values of the parameter
            unique = df[parameter].unique()
            correct = defined[parameter]

            # iterate over the unique values to check them
            for value, correct_value in zip(unique, correct):
                if value is None:
                    self.assertIsNone(correct_value)
                elif np.isnan(value):
                    np.testing.assert_equal(value, correct_value)
                else:
                    self.assertAlmostEqual(value, correct_value)

    def test_concepts(self):
        # set correct number of concepts and variants
        correct_num_variants = 1005
        correct_num_concepts = 550

        # set number of concepts and variants
        num_variants = 0
        num_concepts = 0

        # compute total number of generated concepts + variants
        for _, row in self.configs.df.iterrows():
            # get variantIDs of the concept
            variantIDs = row.concept.variantIDs

            # add to number of variants and concepts
            num_concepts += 1
            num_variants += len(variantIDs)

        # check numbers
        self.assertEqual(num_concepts, correct_num_concepts)
        self.assertEqual(num_variants, correct_num_variants)

    def test_types(self):
        # correct types
        specified = ['RRM', 'CRM', 'RC', 'CC']

        # get types
        types = list(self.configs.df.type.unique())

        self.assertListEqual(types, specified)


def convert(slope):
    if (isinstance(slope, tuple) or isinstance(slope, list)
            or isinstance(slope, np.void)):
        return np.round(np.arctan(slope[0]/slope[1])*180/np.pi, 2)
    else:
        return np.nan


if __name__ == '__main__':
    unittest.main()



import numpy as np
import tkinter as tk
from tkinter.messagebox import showerror
from tkinter.ttk import Separator

from .settings import settings
from .interactive import InteractiveDesign
from .tooltip import ToolTip
from .footer import footer
from ..utils._excel_utils import _convert_string
from ..utils._kwarg_validator import _process_kwargs
from ..utils.exceptions import InputError


def _convert_entry(parameter, entry, vkwargs):
    """ Convert the entry field to the correct format

    Input of an entry field is returned as a string by Tkinter, this
    method converts the input of the entry field to the correct format

    Parameters
    ----------
    parameter : str
        name of the parameter
    entry : str
        input from the entry field
    vkwargs : dict
        dictionary of the valid kwargs for the specified structure

    Returns
    -------
    converted_entry : float, int, tuple, list, str
        input of the entry converted to the correct format
    """
    # get correct format from vkwargs
    values = vkwargs[parameter]

    if not entry:
        converted_entry = None

    elif 'float' in values['Correct']:
        converted_entry = float(entry)

    elif 'int' in values['Correct']:
        converted_entry = int(entry)

    elif 'tuple' in values['Correct'] or 'list' in values['Correct']:
        if 'slope' in parameter:
            mode = 'slope'
        elif 'lambda' in parameter:
            mode = 'lambda'
        else:
            # error
            showerror(
                'Error',
                f'input for {parameter} cannot be given as a tuple or list')

        converted_entry = _convert_string(entry, mode, parameter)

    elif 'str' in values['Correct']:
        converted_entry = entry

    else:
        # pop up with error message
        showerror('Error', f'input for {parameter} is not recognized')

    return converted_entry


class ParameterInput(tk.Frame):
    """ Input page of the app """

    def __init__(self, parent, controller):
        """ See help(ParameterInput) for more info """
        # create the frame
        tk.Frame.__init__(self, parent)

        # set controller as attribute
        self.controller = controller

        # get width and height
        width = settings('lay-out')['width']
        height = settings('lay-out')['height']

        # set height of frames
        top_height = settings('lay-out')['top_height']
        bottom_height = settings('lay-out')['bottom_height']
        center_height = height - top_height - bottom_height

        # set left space
        left_space = width/4

        # create lay out
        top_frame = tk.Frame(self, width=width, height=top_height)
        center_frame = tk.Frame(self, width=width, height=center_height)
        bottom_frame = tk.Frame(self, width=width, height=bottom_height)

        top_frame.pack(side=tk.TOP, fill=tk.X, pady=5)
        center_frame.pack(fill=tk.Y, expand=True)
        bottom_frame.pack(side=tk.BOTTOM, fill=tk.X)

        # set title
        label = tk.Label(
            top_frame, text='Parameter Input',  **settings('title'))
        label.pack()

        # set dict to store the entry field
        entries = {}
        checkboxes = {}

        # first row of the input table
        start_row = 0

        headers = ['Parameter', 'Unit', 'Input', 'Varying']
        for j, header in enumerate(headers):
            label = tk.Label(center_frame, text=header, **settings('label'))
            label.grid(row=start_row, column=j)

        start_row += 1

        # add an entry field for each parameter
        for i, (parameter, value) in enumerate(self.controller.vkwargs.items()):
            if value['Required']:
                required = 'required'
            else:
                required = 'optional'

            format = value['Correct']

            # create label for the parameter
            label = tk.Label(center_frame, text=parameter, **settings('label'))
            label.grid(row=start_row+i, column=0)
            label_tooltip = ToolTip(
                label,
                text=f'Parameter is {required}. Must be formatted as {format}')

            # create label for the unit
            unit = self.controller.vkwargs[parameter]['unit']
            label = tk.Label(center_frame, text=unit, **settings('label'))
            label.grid(row=start_row+i, column=1)

            # create entry fields
            entry = tk.Entry(center_frame)
            entry.grid(row=start_row+i, column=2)
            entry_tooltip = ToolTip(
                entry,
                text=f'Parameter is {required}. Must be formatted as {format}')

            # set default if there is one
            default = value['Default']
            if default is not None:
                entry.insert(i, str(default))

            if not value['Constant']:
                # add checkbox to initiate varying parameter
                var = tk.BooleanVar()
                checkbox = tk.Checkbutton(center_frame, variable=var)
                checkbox.grid(row=start_row+i, column=3)

                # add checkbox to dict
                checkboxes[parameter] = {'var': var, 'check': checkbox}

            # add entry to dict
            entries[parameter] = entry

        # add two empty columns to center input table
        center_frame.grid_columnconfigure(0, weight=1)
        center_frame.grid_columnconfigure(4, weight=1)

        # add lines between the columns
        Separator(center_frame, orient=tk.VERTICAL).grid(
            column=0, row=start_row-1, rowspan=i+2, sticky='nse')
        Separator(center_frame, orient=tk.VERTICAL).grid(
            column=1, row=start_row-1, rowspan=i+2, sticky='nse')

        # add line below the headers
        Separator(center_frame, orient=tk.HORIZONTAL).grid(
            column=0, row=start_row-1, columnspan=4, sticky='ews')

        # add footer
        footer(bottom_frame)

        # add navigation buttons
        nav_settings = settings('navigation')

        start_interactive = tk.Button(
            bottom_frame, text='Start Design', height=nav_settings['height'],
            width=nav_settings['width'],
            command=lambda: self.next_page(entries, checkboxes))
        start_interactive.pack(
            side=tk.RIGHT, padx=nav_settings['padx'], pady=nav_settings['pady'])

        back = tk.Button(
            bottom_frame, text='Back', height=nav_settings['height'],
            width=nav_settings['width'],
            command=lambda: self.controller.home())
        back.pack(
            side=tk.LEFT, padx=nav_settings['padx'], pady=nav_settings['pady'])

    def process_entry(self, entries, varying_parameters):
        """ Process the given entry """
        # set bool to validate the processing
        abort = False

        # output
        output = {}

        # iterate over the entry fields
        for parameter, entry in entries.items():
            if parameter in varying_parameters:
                # given as varying parameter thus pass
                continue
            else:
                # process parameter
                converted_entry = _convert_entry(
                    parameter, entry.get(), self.controller.vkwargs)

                # check if input is valid
                validator = self.controller.vkwargs[parameter]['Validator']
                valid = validator(converted_entry)

                # check if parameter is required
                required = self.controller.vkwargs[parameter]['Required']

                if valid or not required:
                    # add to dict
                    output[parameter] = converted_entry

                else:
                    # error and abort
                    abort = True
                    correct = self.controller.vkwargs[parameter]['Correct']
                    showerror(
                        'InputError',
                        f'{parameter} is incorrectly formatted, '
                        f'must be {correct}')

        # add output parameters to the controller
        self.controller.parameters.update(output)

        return abort

    def process_checkboxes(self, checkboxes):
        """ Process the input from the checkboxes """
        varying_parameters = []

        # iterate over the checkboxes to see if one has been checked
        for parameter, checkbox in checkboxes.items():
            # check if the checkbox was checked
            if checkbox['var'].get():
                # checkbox was checked
                varying_parameters.append(parameter)
            else:
                # checkbox was not checked
                pass

        # return the varying parameters
        return varying_parameters

    def next_page(self, entries, checkboxes):
        """ Process the input and go to the next page """
        # process varying input
        varying_parameters = self.process_checkboxes(checkboxes)

        # process the entry
        abort = self.process_entry(entries, varying_parameters)

        if not abort:
            # check if varying input must be given
            if any(varying_parameters):
                # update attribute for the previous page of InteractiveDesign
                self.controller.previous = 'VaryingInput'

                # go to varying input page
                self.controller.frames[VaryingInput].create_frame(
                    varying_parameters, self.controller.vkwargs)

                self.controller.show_frame(VaryingInput)
            else:
                # update attribute for the previous page of InteractiveDesign
                self.controller.previous = 'ParameterInput'

                # design breakwater for next frame
                self.controller.frames[InteractiveDesign].create_frame()

                # show next frame
                self.controller.show_frame(InteractiveDesign, default=False)


class VaryingInput(tk.Frame):
    """ Optional input frame for giving varying input """

    def __init__(self, parent, controller):
        """ See help(VaryingInput) for more info """
        # create the frame
        tk.Frame.__init__(self, parent)

        # set controller as attribute
        self.controller = controller
        self.first_generation = True

        # get width and height
        width = settings('lay-out')['width']
        height = settings('lay-out')['height']

        # set height of frames
        top_height = settings('lay-out')['top_height']
        bottom_height = settings('lay-out')['bottom_height']
        center_height = height - top_height - bottom_height

        # create lay out
        top_frame = tk.Frame(self, width=width, height=top_height)
        center_frame = tk.Frame(self, width=width, height=center_height)
        bottom_frame = tk.Frame(self, width=width, height=bottom_height)

        top_frame.pack(side=tk.TOP, fill=tk.X, pady=5)
        center_frame.pack(fill=tk.Y, expand=True)
        bottom_frame.pack(side=tk.BOTTOM, fill=tk.X)

        # set frames as attributes
        self.top = top_frame
        self.center = center_frame
        self.bottom = bottom_frame

        # set title of the page
        label = tk.Label(
            self.top, text='Varying Parameter Input', **settings('title'))
        label.pack()

        # add the footer
        footer(self.bottom)

        # create back button
        nav_settings = settings('navigation')

        back = tk.Button(
            self.bottom, text='Back', height=nav_settings['height'],
            width=nav_settings['width'],
            command=lambda: self.controller.show_frame(ParameterInput))
        back.pack(
            side=tk.LEFT, padx=nav_settings['padx'], pady=nav_settings['pady'])

    def create_frame(self, parameters, vkwargs):
        """ Create the frame for giving varying input

        Method to create the frame for when the user wants to specify
        varying parameters.

        Parameters
        ----------
        parameters : list
            list of parameters to variate
        vkwargs : dict
            dict of the valid kwargs for the specified structures
        """
        # check if this is the first time the frame is generated
        if self.first_generation:
            # first time the frame is generated
            # generate entry fields
            self.generate_input(parameters, vkwargs)

            # add start design button
            self.start_interactive_btn()

            # no longer first time the frame is generated
            self.first_generation = False

        else:
            # not the first time
            # clear the widgets of the frame
            for widget in self.center.winfo_children():
                widget.destroy()

            # generate entry fields
            self.generate_input(parameters, vkwargs)

    def start_interactive_btn(self):
        """ Create the start design button """
        nav_settings = settings('navigation')

        # add button to start designing
        start_interactive = tk.Button(
            self.bottom, text='Start Design', height=nav_settings['height'],
            width=nav_settings['width'],
            command=lambda: self.next_page())
        start_interactive.pack(
            side=tk.RIGHT, padx=nav_settings['padx'], pady=nav_settings['pady'])

    def generate_input(self, parameters, vkwargs):
        """ Generate input fields for the varying input

        Method to create the entry fields for the parameters the user
        wants to variate.

        Parameters
        ----------
        parameters : list
            list of parameters to variate
        vkwargs : dict
            dict of the valid kwargs for the specified structures
        """
        # set dict to store the entry field
        entries = {}

        # first row of the input table
        start_row = 0

        headers = ['Parameter', 'Unit', 'Minimum value', 'Maximum value']
        for j, header in enumerate(headers):
            label = tk.Label(self.center, text=header, **settings('label'))
            label.grid(row=start_row, column=j)

        start_row += 1

        # generate the input fields
        for i, parameter in enumerate(parameters):
            # add label for the parameter
            label = tk.Label(self.center, text=parameter, **settings('label'))
            label.grid(row=start_row+i)

            # add label for the unit
            unit = self.controller.vkwargs[parameter]['unit']
            label = tk.Label(self.center, text=unit, **settings('label'))
            label.grid(row=start_row+i, column=1)

            # create entry fields
            entry_min = tk.Entry(self.center)
            entry_max = tk.Entry(self.center)
            entry_min.grid(row=start_row+i, column=2)
            entry_max.grid(row=start_row+i, column=3)

            # add entry to dict
            entries[parameter] = {'min': entry_min, 'max': entry_max}

        # add lines between the columns
        Separator(self.center, orient=tk.VERTICAL).grid(
            column=0, row=start_row-1, rowspan=i+2, sticky='nse')
        Separator(self.center, orient=tk.VERTICAL).grid(
            column=1, row=start_row-1, rowspan=i+2, sticky='nse')

        # add line below the headers
        Separator(self.center, orient=tk.HORIZONTAL).grid(
            column=0, row=start_row-1, columnspan=4, sticky='ews')

        # add two empty columns to center input table
        self.center.grid_columnconfigure(0, weight=1)
        self.center.grid_columnconfigure(4, weight=1)

        # set entries as attribute
        self.entries = entries

    def get_input(self):
        """ Get the input from the entry fields

        Returns
        -------
        dict
            dictionary of the varying parameters with the specified
            minimum and maximum value
        """
        # set bool to validate the processing
        abort = False

        # create dict to store converted entries
        output = {}

        # iterate over the entry fields of the parameters
        for parameter, entry in self.entries.items():
            # create temp dict to store min and max
            temp = {}

            # create variable to store the number of decimals
            num_decimals = 0

            # get the validator and required bool
            validator = self.controller.vkwargs[parameter]['Validator']

            # iterate over the entry field min and max
            for name, value in entry.items():
                # determine the number of decimals in the input
                split = value.get().split('.')

                # check if there are decimals
                if len(split) > 1:
                    # determine number of decimals
                    computed_num_decimals = len(split[-1])
                else:
                    # no decimals
                    computed_num_decimals = 0

                # check if larger than previous num_decimals
                if computed_num_decimals >= num_decimals:
                    num_decimals = computed_num_decimals

                # convert entry field to correct type and store in temp
                converted_entry = _convert_entry(
                    parameter, value.get(), self.controller.vkwargs)

                # check if input is valid
                validator = self.controller.vkwargs[parameter]['Validator']
                valid = validator(converted_entry)

                # check if parameter is valid
                if valid:
                    # check if parameter is a slope parameter
                    if 'slope' in parameter:
                        # set custom number of decimals for degrees
                        num_decimals = 0

                        # convert slope to angles
                        try:
                            converted_entry = np.arctan(
                                converted_entry[0]/converted_entry[1])
                            converted_entry = np.round(
                                converted_entry * 180/np.pi)
                        except IndexError:
                            # parameter not given as a tuple
                            msg = (f'{parameter} must be specified as a '
                                    'tuple (V, H)')
                            showerror('Error', msg)
                            abort = True

                    # add to dict
                    temp[name] = converted_entry

                else:
                    # error and abort
                    abort = True
                    correct = self.controller.vkwargs[parameter]['Correct']
                    showerror(
                        'InputError',
                        f'{name} of {parameter} is incorrectly formatted, '
                        f'must be {correct}')

            # add number of decimals to the dict and temp to output
            temp['resolution'] = num_decimals
            output[parameter] = temp

        return output, abort

    def next_page(self):
        """ Process the input and go to the next page """
        # process varying input
        varying, abort = self.get_input()

        # check if getting input was successful
        if not abort:
            # update parameters with values from the sliders
            for parameter, values in varying.items():
                # get the minimum value
                value = values['min']

                # check if parameter is a slope parameter
                if 'slope' in parameter:
                    # convert angle back to tuple
                    value = np.tan(value*np.pi/180)
                    value = (1, 1/value)

                # add minimum slider value to the parameters
                self.controller.parameters[parameter] = value


            # design breakwater for next frame
            self.controller.frames[InteractiveDesign].create_frame(
                                                            varying=varying)

            # show next frame
            self.controller.show_frame(InteractiveDesign, default=False)



import context
import unittest
import warnings

from breakwater.utils.wave import dispersion
from breakwater.core.scour import scour_protection


class TestScourProtection(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.L = dispersion(T=8, h=15)

    def test_L(self):
        self.assertAlmostEqual(self.L, 81.7897, 4)

    def test_vertical(self):
        w = scour_protection(self.L, slope=None)
        self.assertAlmostEqual(w, 20.4474, 4)

    def test_bounds(self):
        w_lower = scour_protection(self.L, slope=(1,1.75))
        w_upper = scour_protection(self.L, slope=(1,1.2))
        self.assertAlmostEqual(w_lower, 8.1790, 4)
        self.assertAlmostEqual(w_upper, 12.2685, 4)

    def test_out_of_range(self):
        with warnings.catch_warnings(record=True) as w:
            width = scour_protection(self.L, slope=(1,2))

        self.assertAlmostEqual(width, 7.0637, 4)
        self.assertEqual(len(w), 1)
        self.assertEqual(w[0].category, UserWarning)


if __name__ == '__main__':
    unittest.main()



"""
Breakwater Design with Python
"""

__version__ = '1.0'
__author__ = 'S. Winkel'

# import hydraulic conditions
from .core.battjes import BattjesGroenendijk
from .core.goda import goda_wave_heights
from .utils.wave import shoaling_coefficient, dispersion
from .conditions import LimitState

# import materials
from .material import RockGrading, Xbloc, XblocPlus, ConcreteArmour

# import breakwaters
from .design import Configurations
from .rubble import RubbleMound, ConcreteRubbleMound, RockRubbleMound
from .caisson import Caisson

# file loader
from .design import read_breakwaters

# excel input
from .material import read_grading, read_units
from .design import read_configurations
from .utils.input_generator import generate_excel

# interactive design tool (tkinter app)
from .interactive import interactive_design

# import the soil
from .core.soil import Soil



import context
import unittest
import warnings
import numpy as np

from breakwater.core.goda import Goda, goda_wave_heights
from breakwater.utils.wave import shoaling_coefficient


class TestWaveHeights(unittest.TestCase):

    def test_shoaling_coefficient(self):
        # example from Goda (2000), page 78
        Ks = shoaling_coefficient(h=8, T=12, H0=4.5)
        self.assertAlmostEqual(Ks, 1.24, 2)

    def test_wave_heights(self):
        waves = goda_wave_heights(
            h=10.1, d=5.6, Ho=6.3, T=11.4, slope_foreshore=(1,100))
        self.assertAlmostEqual(waves[0], 5.796, 3)
        self.assertAlmostEqual(waves[1], 8.050, 3)


class TestGoda(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # example from Goda (2000), page 140
        with warnings.catch_warnings(record=True) as cls.w:
            cls.goda = Goda(
                Hs=5.796, Hmax=8.050, h=10.1, d=5.6, h_acc=7.1, hc=3.4, Bm=8,
                T=11.4, beta=15*np.pi/180, slope_foreshore=np.arctan(1/100),
                rho=1030, B=15)

    def test_warnings(self):
        self.assertEqual(len(self.w), 1)
        self.assertEqual(self.w[0].category, UserWarning)

    def test_impulsive(self):
        alpha_star = self.goda._impulsive_pressure(Bm=8, Hmax=8.050)
        self.assertAlmostEqual(alpha_star, 0.3223, 4)

    def test_pressures(self):
        self.assertAlmostEqual(self.goda.p1, 97644.84, 2)
        self.assertAlmostEqual(self.goda.p3, 87211.93, 2)
        self.assertAlmostEqual(self.goda.p4, 69674.10, 2)
        self.assertAlmostEqual(self.goda.pu, 65739.31, 2)

    def test_required_mass(self):
        mass = self.goda.required_mass(
            SF_sliding=1.16, SF_turning=2.56, mu=0.6)
        self.assertAlmostEqual(mass/1000, 345.4477, 4)

    def test_bearing_pressure(self):
        pe = self.goda.bearing_pressure(
            Pc=0.72142857, rho_c=2400, rho_fill=1600)
        self.assertAlmostEqual(pe/1000, 291.040, 3)

    def test_mass(self):
        mass = self.goda.mass(Pc=0.8, rho_c=2400, rho_fill=1600)
        self.assertEqual(mass/1000, 352.8)

    def test_centroid(self):
        y = self.goda._pressure_centroid()
        self.assertAlmostEqual(y, 5.155, 3)

    def test_eccentricity(self):
        mass = self.goda.mass(Pc=0.8, rho_c=2400, rho_fill=1600)
        e = self.goda.eccentricity(mass)
        self.assertAlmostEqual(e, 2.049, 3)

    def test_required_width(self):
        with warnings.catch_warnings(record=True) as w:
            B = self.goda.required_width(
                Pc=0.72142857, rho_c=2400, rho_f=1600, rho_w=1030, mu=0.6,
                t=0.5, SF_sliding=1.16, SF_turning=2.56)

            self.assertAlmostEqual(B, 15.200, 3)

        self.assertEqual(len(w), 1)
        self.assertEqual(w[0].category, UserWarning)


if __name__ == '__main__':
    unittest.main()



import setuptools
from breakwater import __version__, __author__

with open('README.md', 'r') as fh:
    long_description = fh.read()

requires = [
    'numpy',
    'scipy',
    'matplotlib',
    'pandas',
    'tabulate',
    ]

setuptools.setup(
    name='breakwater',
    version=__version__,
    author=__author__,
    author_email='pybreakwater@gmail.com',
    description='A package for the conceptual design of breakwaters',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/Sander-w/breakwater/',
    packages=setuptools.find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'License :: Free for non-commercial use',
    ],
    python_requires='>=3.6',
    install_requires = requires,
    keywords='conceptual design hydraulic engineering breakwaters',
    include_package_data=True,
    data_files = [('', ['LICENSE', 'README.md'])],
)



import context
import unittest
import numpy as np

import breakwater.core.toe as toe


class TestRock(unittest.TestCase):

    def test_toe_stability(self):
        Dn50 = toe.toe_stability(Hs=4, h=8, ht=5, Delta=1.6, Nod=0.5)
        self.assertAlmostEqual(Dn50, 0.7411, 4)

    def test_berm_stability(self):
        Dn50 = toe.toe_berm_stability(
            Hs=5.796, T=11.4, d=7.1, Bm=8, Delta=1.6)
        self.assertAlmostEqual(Dn50, 1.58, 2)


if __name__ == '__main__':
    unittest.main()



import sys

class ProgressBar:
    """ Simple progress bar to show progress of a process

    Parameters
    ----------
    number : int
        width of the progress bar
    task : str
        text in front of the progress bar, for instance *Computing*
    max_width : int, optional, default: 50
        maximum width of the progress bar
    """

    def __init__(self, number, task, max_width=50):
        """ See help(ProgressBar) for more info """
        # set attibutes
        self.total = number
        self.task = task
        self.progress = 0

        # set up the progress bar
        self.file = sys.stderr

        # private variable to track if an update should happen
        self._counter = 1
        self._current_number = 0

        # check if maximum width is exceeded by the input
        if number > max_width:
            # if exceeded set width to max allowed width
            self.step = max_width/number
            self.width = max_width
        else:
            # if not exceeded, do nothing
            self.step = 1
            self.width = number

        # print first line
        line = self._line(progress=0)
        print('\r' + line, file=self.file, end='')

    def _line(self, progress):
        """ Print progress bar """
        # compute percentage progress
        bar = '[' + '=' * progress + '.' * (self.width - progress) + ']'
        line = f'{self.task} {bar} {self._current_number}/{self.total}'
        return line

    def next(self):
        """ Add next character to progress bar """
        # create a new line
        self.progress += self.step
        self._current_number += 1

        # check if new update should occur
        if self.progress >= self._counter:
            # update progress bar
            line = self._line(progress=self._counter)
            print('\r' + line, file=self.file, end='')
            self._counter += 1
        else:
            # self._counter-1 because only fraction next to bar is updated
            line = self._line(progress=self._counter-1)
            print('\r' + line, file=self.file, end='')

    def finish(self):
        """ Finish the progress bar """
        # print last part of progress bar
        line = self._line(progress=self._counter)
        print('\r' + line, file=self.file, end='')

        # end progress bar by ending the line
        self.file.write("\n")



import numpy as np

from ..utils.exceptions import InputError


class Soil:
    """ Define soil

    Define the subsoil on which the breakwater is constructed. This
    class is used in the design classes for making the appropriate 
    geotechnical computations.

    .. note::
       It is currently not possible to use several soil layers in the
       design classes. The soil must therefore be a homogeneous soil.

    Parameters
    ----------
    c : float
        cohesion of the soil [kPa]
    phi : float
        internal friction angle [degrees]
    gamma : float, optional, default: None
        volumetric weight of the soil [kN/m³]
    rho : float, optional, default: None
        the density of the soil [kg/m³]
    n : float, optional, default: None
        porosity of the soil

    Attributes
    ----------
    c : float
        cohesion of the soil [kPa]
    phi : float
        internal friction angle [rad]
    gamma : float
        volumetric weight of the soil [kN/m³]
    gamma_sat : float
        saturated volumetric weight of the soil [kN/m³]
    n : float, optional, default: None
        porosity of the soil
    """

    def __init__(self, c, phi, gamma=None, rho=None, n=None):
        """ See help(Soil) for more info """
        # set input as attributes
        self.c = c
        self.phi = phi * np.pi/180

        # set porosity as attribute
        self.n = n

        # set volumetric weight as attribute
        if gamma is None:
            # check if rho has been given
            if rho is None or n is None:
                raise InputError(
                    ('When the volumetric is not specified, rho and n must '
                     'be specified'))
            else:
                # compute with porosity
                self.gamma = ((1-n)*rho*9.81)/1000
        else:
            # given as input
            self.gamma = gamma

        # set gamma_sat attribute
        self.gamma_sat = None

    def saturated_weight(self, gamma_sat=None, rho_w=None):
        """ Method to add the saturated weight

        Method to add the properties of the soil if it is saturated.
        The saturated soil properties can be added by setting gamma_sat
        or by specifing rho_w. In case of the latter the saturated weight
        is computed with the porosity.

        Parameters
        ----------
        gamma_sat : float, optional, default: None
            saturated volumetric weight of the soil [kN/m³]
        rho_w : float, optional, default: None
            density of water [kg/m³]

        Raises
        ------
        InputError
            if the porosity is not specified and the saturated
            volumetric weight is computed with rho_w
        """
        # check if gamma_sat has been given
        if gamma_sat is not None:
            # set with gamma_sat
            self.gamma_sat = gamma_sat

        else:
            # compute with porosity, check if porosity has been specified
            if self.n is not None:
                # compute saturated volumetric weight
                self.gamma_sat = self.gamma + (self.n*rho_w*9.81)/1000

            else:
                # no porosity has been specified
                raise InputError(
                    ('porosity must be specified when computing the '
                     'saturated volumetric weight'))

    def brinch_hansen(self, p, t, B, L, q, rho_w=None, sat=True):
        """ Brinch-Hansen

        implementation of the Brinch Hansen equation to determine
        the bearing capacity of the soil per unit length. The equation
        is given by (Brinch Hansen, 1970):

        .. math::
           p=i_{c} s_{c} c N_{c}+i_{q} s_{q} q N_{q}+i_{\\gamma}
           s_{\\gamma} \\frac{1}{2} \\gamma B N_{\\gamma}

        In which :math:`i` and :math:`s` are the inclination and shape
        factor, and :math:`N` are dimensionless constants. Note that
        compared to the original equation the depth, base and ground
        inclination factors have been omitted. The latter two because
        the assumption is made that the foundation is never constructed
        at an angle, and the depth factor because the foundation is not
        placed in the soil but on top of the bottom.

        .. note
           the soil is assumed to be homogeneous

        Parameters
        ----------
        t : float
            horizontal stress [kPa]
        p : float
            vertical stress [kPa]
        q : float
            overburden [kPa]
        B : float
            width of the structure [m]
        L : float
            length of the structure [m], set to None if the structure is
            a long structure and the shape factors can be neglected.
        rho_w : float, optional, default: None
            density of water, by default the soil is not submerged [kg/m³]
        sat : bool, optional, default: True
            True if the saturated volumetric weight of the soil must be
            used, False if the dry volumetric weight of the soil must be
            used.

        Returns
        -------
        p : float
            bearing capacity of the foundation [kPa]
        """
        # check density to use
        if sat:
            # check if saturated volumetric weight is given
            if self.gamma_sat is not None:
                # compute effective volumetric weight
                gam = self.gamma_sat - (rho_w*9.81)/1000
            else:
                # not given, but is needed
                raise InputError(
                    ('saturated volumetric weight has not been defined, use '
                     'saturated_weight() to define saturated weight'))
        else:
            # use dry volumetric weight
            gam = self.gamma

        # compute dimensionless constants
        N_q = ((1 + np.sin(self.phi))/(1 - np.sin(self.phi))
                * np.exp(np.pi * np.tan(self.phi)))
        N_c = (N_q - 1)/np.tan(self.phi)
        N_gam = 2 * (N_q - 1) * np.tan(self.phi)

        # compute inclination factors
        i_c = 1 - t / (self.c + p * np.tan(self.phi))
        i_q = i_c**2
        i_gam = i_c**3

        # check if structure is a long structure
        if L is None:
            # shape factors can be neglected, set all shape factors to 1
            s_c, s_q, s_gam = 1, 1, 1
        else:
            # compute shape factors
            s_c = 1 + 0.2 * B/L
            s_q = 1 + np.sin(self.phi) * B/L
            s_gam = 1 - 0.3 * B/L

        # compute capacity of the soil
        p = i_c*s_c*self.c*N_c + i_q*s_q*q*N_q + 0.5*i_gam*s_gam*gam*B*N_gam

        return p



import matplotlib
import numpy as np
import tkinter as tk
from tkinter.ttk import Treeview
from tkinter.messagebox import showerror, showinfo
from warnings import catch_warnings
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
from matplotlib.figure import Figure

from .settings import settings, size
from .footer import footer
from ..rubble import RockRubbleMound, ConcreteRubbleMound
from ..caisson import Caisson
from ..utils.exceptions import RockGradingError, ArmourUnitsError
from ..utils._kwarg_validator import _RM_vkwargs, _C_vkwargs
from ..utils.cost import cost_influence


class InteractiveDesign(tk.Frame):
    """ Interactive design page of the app """

    def __init__(self, parent, controller):
        """ See help(ParameterInput) for more info """
        tk.Frame.__init__(self, parent)

        # set controller as attribute
        self.controller = controller

        # set attribute for the sliders
        self.sliders = {}

        # create attribute to store varying parameters of the bw_types
        self.varying = {'RM': {}, 'C': {}}

        # set attribute to track button generation
        self.buttons = {'update': False, 'influence': False}

        # set attribute for the limits of the xaxis
        self.xlim = None

        # set variables
        width, height = size(
            self.controller.to_design, 'interactive',
            self.controller.num_concepts)

        # determine dimensions of the frames
        top_height = settings('lay-out')['top_height']
        bottom_height = settings('lay-out')['bottom_height']

        # set top and bottom frame
        top_frame = tk.Frame(self, width=width, height=top_height)
        self.bottom_frame = tk.Frame(self, width=width, height=bottom_height)

        # pack frames
        top_frame.pack(side=tk.TOP, fill=tk.X, pady=5)
        self.bottom_frame.pack(side=tk.BOTTOM, fill=tk.X)

        # generate design frames
        self._generate_design_frames(width, height)

        # add the title of the page
        label = tk.Label(
            top_frame, text='Interactive Design', **settings('title'))
        label.pack()

        # add the footer
        footer(self.bottom_frame)

        # add back button
        nav_settings = settings('navigation')

        back = tk.Button(
            self.bottom_frame, text='Back', height=nav_settings['height'],
            width=nav_settings['width'], command=lambda: self.back())
        back.pack(
            side=tk.LEFT, padx=nav_settings['padx'], pady=nav_settings['pady'])

    def _generate_design_frames(self, width, height):
        """ Generate the frames for the designs

        Parameters
        ----------
        width, height : int
            width and height of the super frame in pixels
        """
        # set attribute for the design frames and sliders
        self.design_frames = {}
        self.slider_frames = {}

        # frame variables
        design_width, design_height = size(
            self.controller.to_design, 'design', self.controller.num_concepts)
        info_width = settings('interactive')['info_width']
        slider_width = width - design_width - info_width

        # generate RM design frames
        for bw_type, values in self.controller.to_design.items():
            # check the number of variants for this type
            num = values['num']
            if num > 0:
                # create parent frame
                parent = tk.Frame(self, width=width, height=design_height*num)
                parent.pack_propagate(False)
                parent.pack()

                # add frame for the sliders
                sliders = tk.Frame(
                    parent, width=slider_width, height=design_height*num)
                sliders.pack_propagate(False)
                sliders.pack(side=tk.LEFT)

                # create parent for the design frames
                design_parent = tk.Frame(
                    parent, width=design_width+info_width,
                    height=design_height*num)
                design_parent.pack_propagate(False)
                design_parent.pack(side=tk.RIGHT)

                # create dict to store the design frames of this type
                design_frames = {}

                # remove num from values
                del values['num']

                # create frames for the bw_type
                for structure, active in values.items():
                    # check if structure is active
                    if active:
                        # structure is active thus create the required frames
                        # create parent for the types
                        type_parent = tk.Frame(
                            design_parent, width=design_width+info_width,
                            height=design_height)
                        type_parent.pack_propagate(False)
                        type_parent.pack()

                        # create plot frame
                        plot_frame = tk.Frame(
                            type_parent, width=design_width, height=design_height)
                        plot_frame.pack_propagate(False)
                        plot_frame.pack(side=tk.LEFT, padx=0, pady=0)

                        # create info frame
                        info_frame = tk.Frame(
                            type_parent, width=info_width, height=design_height)
                        info_frame.pack(side=tk.RIGHT, padx=5)

                        # get the name of the structure and add frames to dict
                        design_frames[structure] = {
                            'plot': plot_frame, 'info': info_frame}

                # add frames to attribute
                self.design_frames[bw_type] = design_frames
                self.slider_frames[bw_type] = sliders

    def _display_warning(self, w, structure, show_warnings):
        """ display the warning messages encountered during the design

        Parameters
        ----------
        w : list
            list of warnings
        structure : str
            name of the structure to design
        show_warnings : bool
            True if warnings must be displayed, False if not
        """
        # check if there are warnings
        if len(w) > 0 and show_warnings:
            # there are warnings
            msg = 'Warnings: \n'
            for warning in w:
                msg += f'- {warning.message} \n'

            showinfo(f'{structure} Warning', msg)

        else:
            # no warnings
            pass

    def _format_input(self, structure):
        """ Format the input for the design class

        Function to format the parameters to the correct format for the
        design classes. This means that for the RC the BermMaterial
        argument is set to the Grading.

        Parameters
        ----------
        structure : str
            name of the structure to design

        Returns
        -------
        dict
            with the parameters correctly formatted
        """
        # check structure
        if structure == 'RRM' or structure == 'CRM':
            # no formatting required
            return self.controller.parameters

        elif structure == 'RC' or structure == 'CC':
            # get the parameters
            output = self.controller.parameters.copy()

            # caisson structure, remove layers rock and units
            if 'layers_rock' in output.keys():
                del output['layers_rock']

            if 'layers_rock' in output.keys():
                del output['layers_units']

            if structure == 'RC':
                if 'BermMaterial' in output.keys():
                    del output['BermMaterial']

            return output

    def back(self):
        """ Back to the previous page """
        # remove the xlims of the previous plot
        self.xlim = None

        # clear the buttons in the bottom frame
        to_delete = ['Update Design', 'Cost influence']
        for widget in self.bottom_frame.winfo_children():
            if widget.cget('text') in to_delete:
                widget.destroy()

        self.buttons = {'update': False, 'influence': False}

        # clear the figures
        for bw_type, structures in self.design_frames.items():
            for structure, frames in structures.items():
                # destroy all widgets
                for widget in frames['plot'].winfo_children():
                    widget.destroy()

                for widget in frames['info'].winfo_children():
                    widget.destroy()

                for widget in self.slider_frames[bw_type].winfo_children():
                    widget.destroy()

                # delete the designs
                if hasattr(self, f'{structure}_old'):
                    delattr(self, f'{structure}_old')

                if hasattr(self, f'{structure}_new'):
                    delattr(self, f'{structure}_new')

        # destroy the sliders
        for bw_type, frame in self.slider_frames.items():
            for widget in frame.winfo_children():
                widget.destroy()

        # go to the previous page
        self.controller.back(self.controller.previous)

    def create_frame(self, varying=None):
        """ Create frame

        Method to create the page when the ParameterInput, and
        VaryingInput (if this is specified), is given.

        Parameters
        ----------
        varying : dict, optional, default: None
            dictionary of the varying parameters with the specified
            minimum and maximum value
        """
        # iterate over the design frames to plot the design and show info
        for bw_type, structures in self.design_frames.items():
            # check if varying parameters have been given
            if varying is not None:
                # varying parameters, thus generate sliders
                self.generate_sliders(
                    varying, self.slider_frames[bw_type], bw_type)

                # check if cost have been specified
                if self.controller.cost is not None:
                    # create cost button
                    nav_settings = settings('navigation')

                    if not self.buttons['influence']:
                        cost_button = tk.Button(
                            self.bottom_frame, text='Cost influence',
                            height=nav_settings['height'],
                            width=nav_settings['width'],
                            command=lambda: self.cost_influence())
                        cost_button.pack(
                            side=tk.RIGHT, padx=nav_settings['padx'],
                            pady=nav_settings['pady'])

                        self.buttons['influence'] = True

            else:
                # no varying input
                no_sliders(self.slider_frames[bw_type])

            # iterate over the structures of the type
            for structure, frames in structures.items():
                # show the breakwater
                self.show_breakwater(structure, frames)

    def generate_sliders(self, varying, frame, bw_type):
        """ Add sliders for the varying parameters

        Method to create the sliders for interactive design

        Parameters
        ----------
        varying : dict
            dictionary of the varying parameters with the specified
            minimum and maximum value
        frame : tk.Frame
            frame on which the sliders must be placed
        bw_type : str
            type of breakwater
        """
        # determine the width of a slider
        width = settings('interactive')['width']
        design_width = settings('interactive')['design_width']
        info_width = settings('interactive')['info_width']
        frame_width = width - design_width - info_width

        width = _slider_width(
            parameters=varying.keys(), frame_width=frame_width)

        # create bool to check if a slider has been placed
        placed_slider = False

        # get the vkwargs of the structure
        if bw_type == 'RM':
            vkwargs = _RM_vkwargs(type='both')
        elif bw_type == 'C':
            vkwargs = _C_vkwargs(type='both')
        else:
            raise NotImplementedError(f'{bw_type} has not been implemented')

        # create frame for the sliders
        parent = tk.Frame(
            frame, width=frame.winfo_reqwidth(), height=frame.winfo_reqheight())
        parent.place(anchor=tk.CENTER, relx=.5, rely=.5)

        # iterate over the varying parameters
        for parameter, values in varying.items():
            # check if parameter is valid for the given structure
            if parameter in vkwargs.keys():
                # make slider and add slider to attribute
                self.sliders[parameter] = make_slider(
                    parameter, values, parent, width)

                # change bool
                placed_slider = True

                # add to dict
                self.varying[bw_type][parameter] = varying[parameter]

        # check if a slider has been placed
        if not placed_slider:
            # no slider, add label
            no_sliders(frame)

        # create update button
        nav_settings = settings('navigation')

        if not self.buttons['update']:
            update_button = tk.Button(
                self.bottom_frame, text='Update Design',
                height=nav_settings['height'], width=nav_settings['width'],
                command=lambda: self.update())
            update_button.pack(
                side=tk.RIGHT, padx=nav_settings['padx'],
                pady=nav_settings['pady'])

            self.buttons['update'] = True

    def update(self):
        """ Update design with new varying parameters

        Method to update the design with the new parameters from the
        sliders
        """
        # iterate over the sliders
        for parameter, value in self.sliders.items():
            # get the value of the slider
            value = value.get()

            # check if parameter is a slope parameter
            if 'slope' in parameter:
                # convert angle back to tuple
                value = np.tan(value*np.pi/180)
                value = (1, 1/value)

            self.controller.parameters[parameter] = value

        # remove the xlims of the previous plot
        self.xlim = None

        # iterate over the design frames to plot the design and show info
        for bw_type, structures in self.design_frames.items():
            # iterate over the structures of the type
            for structure, frames in structures.items():
                # delete the old plot
                for widget in frames['plot'].winfo_children():
                    widget.destroy()
                for widget in frames['info'].winfo_children():
                    widget.destroy()

                # show the breakwater
                self.show_breakwater(structure, frames)

    def show_breakwater(self, structure, frames):
        """ Method to generate the plot and info

        Parameters
        ----------
        structure : str
            name of the structure to design
        frames : dict
            dictionary of tk.Frames
        """
        # design the breakwater
        self.design(structure, show_warnings=self.controller.display_warnings)

        # compute the cost and get cheapest variant
        cost = self.compute_cost(structure)

        if cost is not None:
            min_variant = min(cost, key=cost.get)
            price = cost[min_variant]
        else:
            min_variant = 'a'
            price = None

        # add the plot the frame
        self.add_plot(structure, frames['plot'], min_variant)

        # add the info of the design
        self.add_info(structure, frames['info'], min_variant, price)

    def design(self, structure, show_warnings, show_error=True):
        """ Design a breakwater

        Parameters
        ----------
        structure : str
            name of the structure to design
        show_warnings : bool
            True if warnings must be displayed, False if not
        show_error : bool, optional, default: True
            True if errors must be shown, False if not
        """
        # check if a design has already been made
        if hasattr(self, f'{structure}_old'):
            # not the first design
            # set previous design as old design
            previous_design = getattr(self, f'{structure}_new')
            setattr(self, f'{structure}_old', previous_design)

        else:
            # first design, set attributes for old design
            setattr(self, f'{structure}_old', None)

        # format input parameters
        input = self._format_input(structure)

        if structure == 'RRM':
            # design rubble mound breakwater with rock
            try:
                with catch_warnings(record=True) as w:
                    self.RRM_new = RockRubbleMound(**input)

                self._display_warning(w, 'RRM', show_warnings)

            except RockGradingError as e:
                if show_error:
                    showerror('RockGradingError', f'Computed {e}')
                delattr(self, 'RRM_old')

        elif structure == 'CRM':
            # design rubble mound breakwater with armour units
            try:
                with catch_warnings(record=True) as w:
                    self.CRM_new = ConcreteRubbleMound(**input)

                self._display_warning(w, 'CRM', show_warnings)

            except ArmourUnitsError as e:
                if show_error:
                    showerror('ArmourUnitsError', f'Computed {e}')
                delattr(self, 'CRM_old')

        elif structure == 'RC':
            # design caisson breakwater with rock as armour for the foundation
            try:
                with catch_warnings(record=True) as w:
                    self.RC_new = Caisson(
                        layers=self.controller.parameters['layers_rock'],
                        BermMaterial=self.controller.parameters['Grading'],
                        **input)

                self._display_warning(w, 'RC', show_warnings)

            except RockGradingError as e:
                if show_error:
                    showerror('RockGradingError', f'Computed {e}')
                delattr(self, 'RC_old')

        elif structure == 'CC':
            # design caisson breakwater with armour units as armour
            try:
                with catch_warnings(record=True) as w:
                    self.CC_new = Caisson(
                        layers=self.controller.parameters['layers_units'],
                        **input)

                self._display_warning(w, 'CC', show_warnings)

            except ArmourUnitsError as e:
                if show_error:
                    showerror('ArmourUnitsError', f'Computed {e}')
                delattr(self, 'CC_old')

        else:
            # invalid structure
            raise NotImplementedError(f'{structure} has not been implemented')

    def compute_cost(self, structure, output='variant'):
        """ Determine the most expensive variant

        Parameters
        ----------
        structure : str
            name of the structure to design
        output : {variant, layer, average}, optional, default: variant
            format of the output dict, variant returns the total cost
            of each variant, layer the cost of each layer for each
            variant and average returns the average cost.
        Returns
        -------
        dict
            the cost of each variant
        """
        # get the attribute
        design = getattr(self, f'{structure}_new')

        # check if cost have been specified
        if self.controller.cost is not None:
            # cost have been added
            if structure == 'RRM':
                # rubble mound structure with rock as armour
                cost = design.cost(
                    *design.variantIDs,
                    core_price=self.controller.cost['core_price'],
                    transport_cost=self.controller.cost['transport_price'],
                    output=output)

            elif structure == 'CRM':
                # rubble mound structure with ArmourUnits as armour
                cost = design.cost(
                    *design.variantIDs,
                    core_price=self.controller.cost['core_price'],
                    unit_price=self.controller.cost['unit_price'],
                    transport_cost=self.controller.cost['transport_price'],
                    output=output)

            elif structure == 'RC' or structure == 'CC':
                # get rent of dry dock and length of the breakwater
                rent = self.controller.cost['dry_dock']
                length = self.controller.cost['length']

                # check if dry dock investment must be added
                if rent is not None and length is not None:
                    # add rent of dry dock
                    design.dry_dock(rent, length)

                # check if rock armour layer
                if structure == 'RC':
                    # caisson structure with rock as armour of the foundation
                    cost = design.cost(
                        *design.variantIDs,
                        concrete_price=self.controller.cost['concrete_price'],
                        fill_price=self.controller.cost['fill_price'],
                        output=output)

                else:
                    # caisson structure with units as armour of the foundation
                    cost = design.cost(
                        *design.variantIDs,
                        concrete_price=self.controller.cost['concrete_price'],
                        fill_price=self.controller.cost['fill_price'],
                        unit_price=self.controller.cost['unit_price'],
                        output=output)

            else:
                # invalid structure
                raise NotImplementedError(
                    f'{structure} has not been implemented')

            # return the computed cost
            return cost

        else:
            return None

    def cost_influence(self, num=50):
        """ Plot the influence of the varying parameters on the cost """
        # make dict to store designs
        store_designs = {}

        # make variable to store current parameters
        parameters = self.controller.parameters.copy()

        # make dict to store the cost data and values
        lines = {}

        # iterate over the structures to design
        for bw_type, structures in self.controller.to_design.items():
            # get varying parameters of this breakwater type
            varying = self.varying[bw_type]
            # iterate over the structures
            for structure, active in structures.items():
                # check if structure is active
                if active:
                    # get current old and new design, and store in dict
                    store_designs[structure] = {
                        'old': getattr(self, f'{structure}_old'),
                        'new': getattr(self, f'{structure}_new')}

                    # compute influence of varying parameter on the cost
                    # iterate over the varying parameters
                    for parameter, values in varying.items():
                        # reset parameters back to original parameters
                        self.controller.parameters = parameters.copy()

                        # make linspace
                        varying = np.linspace(
                            values['min'], values['max'], num)

                        # add to line data
                        lines[parameter] = {'values': varying, 'cost': []}

                        # iterate over the values
                        for value in varying:
                            # check if parameter is a slope parameter
                            if 'slope' in parameter:
                                # convert angle back to tuple
                                value = np.tan(value*np.pi/180)
                                value = (1, 1/value)

                            self.controller.parameters[parameter] = value

                            # update design and get cost
                            self.design(
                                structure, show_warnings=False,
                                show_error=False)

                            with catch_warnings(record=True) as w:
                                cost = self.compute_cost(
                                    structure, output='average')

                            # add to line
                            lines[parameter]['cost'].append(
                                list(cost.values())[0])

        # create figure
        cost_influence(lines)

        # set design and parameters back to original ones
        for structure, concepts in store_designs.items():
            for key, concept in concepts.items():
                setattr(self, f'{structure}_{key}', concepts[key])

        self.controller.parameters = parameters

    def add_info(self, structure, frame, variantID, price):
        """ Add info of the plotted structure

        Parameters
        ----------
        structure : str
            name of the structure to design
        frame : tk.Frame
            frame to add the info to
        variantID : str
            identifier of the variant of which the info must be displayed
        price : float
            price of the specified variant
        """
        # get the dimensions of the frame
        param_width = 150
        value_width = settings('interactive')['info_width'] - param_width - 10

        # get the font
        font = settings('font')['info']

        # get the concept breakwater and variant
        concept = getattr(self, f'{structure}_new')
        variant = concept.get_variant(variantID)

        # create the treeview
        tree = Treeview(frame)

        # add columns
        tree['columns'] = ('#1')
        tree.column('#0', width=param_width, stretch=tk.NO)
        tree.heading('#0',text='Parameter', anchor=tk.W)
        tree.column('#1', width=value_width, stretch=tk.NO)
        tree.heading('#1', text='Value', anchor=tk.W)

        # check if a price has been specified
        if price is not None:
            # add cost
            tree.insert('', 1, iid=1, text='Cost', values=(np.round(price,2)))

            # set row
            row = 2

        else:
            # no cost, set row to first one
            row = 1

        # check if RubbleMound structure
        if structure == 'RRM' or structure == 'CRM':
            # parameters to show
            includes = ['computed Dn50', 'class']

            # add Rc
            tree.insert(
                '', row, iid=row, text='Rc',
                values=(f'{np.round(concept.Rc, 3)}m'))

            # increment row
            row += 1

            # add the info
            for layer, dimensions in variant.items():
                # add to first column
                header = tree.insert(
                    '', row, iid=row, text=layer.capitalize(), values=(''))

                # increment row
                row += 1

                # iterate over the dimensions of the layer
                for parameter, value in dimensions.items():
                    # check if value is an int or float
                    if isinstance(value, int) or isinstance(value, float):
                        if layer == 'armour' and parameter == 'class':
                            if structure == 'CRM':
                                SUB = str.maketrans('3', '³')
                                unit = 'm3'.translate(SUB)
                            else:
                                unit = 'm'
                        else:
                            unit = 'm'
                        # round the dimension to 3 decimals
                        value = f'{np.round(value, 3)}{unit}'

                    if parameter in includes:
                        # parameter and value
                        tree.insert(
                            header, 'end', iid=row,
                            text=parameter.capitalize(), values=(value))

                        # increment the row
                        row += 1

            # pack the tree to the frame
            tree.pack(side=tk.TOP, fill=tk.Y)

        # check if Caisson breakwater
        elif structure == 'RC' or structure == 'CC':
            # parameters to show
            includes = ['computed Dn50', 'class', 'd', 'B', 'h_acc']
            no_capital = ['h_acc', 'd']

            # add Rc
            Rc = variant['caisson']['Rc']
            tree.insert(
                '', row, iid=row, text='Rc', values=(f'{np.round(Rc, 3)}m'))

            # increment row
            row += 1

            # add the info
            for layer, dimensions in variant.items():
                # add to first column
                header = tree.insert(
                    '', row, iid=row, text=layer.capitalize(), values=(''))

                # increment row
                row += 1

                # iterate over the dimensions of the layer
                for parameter, value in dimensions.items():
                    # check if value is an int or float
                    if isinstance(value, int) or isinstance(value, float):
                        if layer == 'armour' and parameter == 'class':
                            if structure == 'CC':
                                SUB = str.maketrans('3', '³')
                                unit = 'm3'.translate(SUB)
                            else:
                                unit = 'm'
                        else:
                            unit = 'm'

                        # round the dimension to 3 decimals
                        value = f'{np.round(value, 3)}{unit}'

                    if parameter in includes:
                        if parameter not in no_capital:
                            parameter = parameter.capitalize()
                        # parameter and value
                        tree.insert(
                            header, 'end', iid=row, text=parameter,
                            values=(value))

                        # increment the row
                        row += 1

            # add geotechnical parameters if computed
            if self.controller.parameters['Soil'] is not None:
                # define the stresses
                stresses = ['p', 'p_cap']

                # add header of the geotechnical parameters
                geo = tree.insert(
                    '', row, iid=row, text='Geotechnical', values=(''))

                # increment row
                row += 1

                # add values
                for parameter, value in concept.geotechnical.items():
                    # check if parameter is a stess
                    if parameter in stresses:
                        # stress, round value and add unit
                        value = f'{np.round(value, 2)}kPa'
                    else:
                        # only round value
                        value = f'{np.round(value, 2)}'

                    # add to the tree and increment row
                    tree.insert(
                        geo, 'end', iid=row, text=parameter, values=(value))
                    row += 1

            # pack the tree to the frame
            tree.pack(side=tk.TOP, fill=tk.Y)

        else:
            # other type
            raise NotImplementedError(f'{structure} has not been implemented')

    def add_plot(self, structure, frame, variantID):
        """ Plot the cross section of the specified breakwater

        Parameters
        ----------
        structure : str
            name of the structure to design
        frame : tk.Frame
            frame to add the plot to
        variantID : str
            identifier of the variant to plot
        """
        # plot variables
        dpi = settings('interactive')['dpi']
        design_width, design_height = size(
            self.controller.to_design, 'design', self.controller.num_concepts)
        tick_fontsize = 6

        # create the figure
        fig = Figure(
            figsize=(design_width/dpi, design_height/dpi), dpi=dpi,
            facecolor='#f0f0f0')

        # add subplot to the figure
        ax = fig.add_subplot(111)

        # get the previous design
        old_design = getattr(self, f'{structure}_old')

        # check if old design must be plotted
        if old_design is not None:
            # get coordinates of the old design
            coordinates = old_design._layers(variantID)

            # iterate over the layers
            for layer, lines in coordinates.items():
                # plot each layer
                ax.plot(
                    lines['x'], lines['y'], color='grey', ls='--', lw=1,
                    zorder=1)

        # get the new design
        new_design = getattr(self, f'{structure}_new')

        # get coordinates of the new design
        coordinates = new_design._layers(variantID)

        # iterate over the layers
        for layer, lines in coordinates.items():
            # plot each layer
            ax.plot(lines['x'], lines['y'], color='k', zorder=20, lw=1)

        # set xlim
        xmin = ax.get_xlim()[0]
        xmax = ax.get_xlim()[1]

        if self.xlim is None:
            self.xlim = {'xmin': xmin, 'xmax': xmax}
        else:
            if np.abs(self.xlim['xmin']) > np.abs(xmin):
                xmin = self.xlim['xmin']

            if np.abs(self.xlim['xmax']) > np.abs(xmax):
                xmax = self.xlim['xmax']

        ax.set_xlim(xmin, xmax)

        # add bottom and wlev
        if structure == 'RRM' or structure == 'CRM':
            # rubble mound structure
            x_wlev = (0.5*new_design._input_arguments['B']
                      + new_design._input_arguments['slope'][1]*new_design.Rc
                            /new_design._input_arguments['slope'][0])

            wlev = new_design._LimitStates[new_design._state_overtopping].h

        elif structure == 'RC' or structure == 'CC':
            # vertical structure
            x_wlev = np.max(coordinates['caisson']['x'])
            wlev = new_design._LimitStates[
                        new_design.structure['caisson']['state_overtop']].h

        ax.hlines(
            y=wlev, xmin=xmin, xmax=-x_wlev, color='dodgerblue', lw=1)
        ax.hlines(
            y=wlev, xmin=x_wlev, xmax=xmax, color='dodgerblue', lw=1)
        ax.hlines(y=0, xmin=xmin, xmax=xmax, color='peru', lw=1, zorder=5)

        # format the ticks
        for xtick in ax.get_xticklabels():
            xtick.set_fontsize(tick_fontsize)

        for ytick in ax.get_yticklabels():
            ytick.set_fontsize(tick_fontsize)

        # format the plot with equal aspect and removing whitespace
        ax.set_aspect('equal', adjustable='box')
        ax.grid()
        fig.subplots_adjust(left=0.08, bottom=0.08, right=0.95, top=1)

        # create canvas for the figure
        canvas = FigureCanvasTkAgg(fig, frame)

        # show canvas and pack canvas
        canvas.draw()
        canvas.get_tk_widget().pack(expand=False)

def no_sliders(frame):
    """ Place a label for when no sliders have been added

    Parameters
    ----------
    frame : tk.Frame
        frame on which to pack the label
    """
    label = tk.Label(
        frame,
        text='No varying input has been specified\n Sliders not available')
    label.pack(fill=tk.Y, expand=True)

def _slider_width(parameters, frame_width):
    """ Determine the width of a slider

    Parameters
    ----------
    parameters : list
        list of parameters
    frame_width : int
        width of the frame in pixels

    Returns
    -------
    int
        required width of the slider to fill the frame
    """
    # get the longest parameter
    max_len = 0

    for parameter in parameters:
        length = len(parameter)
        if length >= max_len:
            max_len = length

    # width of a character is pixels
    char_width = 7

    # compute width of the slider
    width = frame_width - max_len*char_width - 5

    return width

def make_slider(parameter, values, frame, slider_width):
    """ Make a slider and label

    Function to make a slider and label and place them on the specified
    frame

    Parameters
    ----------
    parameter : str
        name of the parameter
    values : dict
        dictionary with the minimum and maximum value of the slider
    frame : tk.Frame
        frame on which to place the label and slider
    slider_width : int
        required width of a slider
    """
    # height of a slider
    slider_height = 42

    # create frame for the slider and label
    parent = tk.Frame(
        frame, width=frame.winfo_reqwidth(), height=slider_height)
    parent.pack_propagate(False)
    parent.pack(side=tk.TOP)

    # create frame for the label
    label_frame = tk.Frame(
        parent, width=frame.winfo_reqwidth()-slider_width,
        height=slider_height)
    label_frame.pack_propagate(False)
    label_frame.pack(side=tk.LEFT)

    # create label
    label = tk.Label(label_frame, text=parameter, padx=0)
    label.pack(side=tk.BOTTOM)

    # create frame for the slider
    slider_frame = tk.Frame(parent, width=slider_width, height=slider_height)
    slider_frame.pack_propagate(False)
    slider_frame.pack(side=tk.RIGHT)

    # determine the resolution of the slider
    resolution = 1/(10**(values['resolution']+1))

    # create slider
    slider = tk.Scale(
        slider_frame, from_=values['min'], to=values['max'],
        orient=tk.HORIZONTAL, resolution=resolution, length=slider_width)
    slider.pack()

    # return the slider
    return slider



import context
import unittest
import numpy as np

import breakwater.core.stability as stability


class TestRock(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.slope = np.arctan(1/3)

    def test_xi_cr(self):
        # deep water model constants, Rock Manual (2007) Box 5.13
        xi_cr_deep = stability.xi_critical(
            Cpl=6.2, Cs=1, P=0.4, alpha=self.slope)
        self.assertAlmostEqual(xi_cr_deep, 3.0, 1)

    def test_vandermeer_deep(self):
        # plunging conditons, Rock Manual (2007) Box 5.13
        Dn50 = stability.vandermeer_deep(
            Hs=5, Delta=1.6, P=0.4, Sd=5, N=2100, xi_m=1.85, alpha=self.slope,
            safety=0)
        self.assertAlmostEqual(Dn50, 1.26, 2)
        # plunging conditions + safety, Rock Manual (2007) Box 5.14
        Dn50 = stability.vandermeer_deep(
            Hs=5, Delta=1.6, P=0.4, Sd=5, N=2100, xi_m=1.85, alpha=self.slope,
            safety=1.75)
        self.assertAlmostEqual(Dn50, 1.42, 2)
        # test surging conditions
        Dn50 = stability.vandermeer_deep(
            Hs=4, Delta=1.6, P=0.4, Sd=5, N=2100, xi_m=4.46,
            alpha=np.arctan(1/1.5),safety=0)
        self.assertAlmostEqual(Dn50, 1.5519, 4)

    def test_vandermeer_shallow(self):
        # plunging conditions, Rock Manual (2007) Box 5.15 (with errata)
        Dn50 = stability.vandermeer_shallow(
            Hs=4, H2=4.95, Delta=1.6, P=0.4, Sd=2, N=2273, xi_s_min_1=2.39,
            alpha=self.slope, safety=0)
        self.assertAlmostEqual(Dn50, 1.27, 2)
        # surging conditons
        Dn50 = stability.vandermeer_shallow(
            Hs=3.5, H2=4.2, Delta=1.6, P=0.4, Sd=2, N=2273, xi_s_min_1=4.01,
            alpha=np.arctan(1/2), safety=0)
        self.assertAlmostEqual(Dn50, 1.3713, 4)

    def test_hudson(self):
        # Checked with values from Xbloc guidelines, table 1
        Dn = stability.hudson(H=5.01, Kd=16, Delta=1.33, alpha=np.arctan(3/4))
        self.assertAlmostEqual(Dn**3, 2.5, 1)


if __name__ == '__main__':
    unittest.main()
